setMethod("Ops",
          signature(e1="externalVectorWithStorage"),
          function(e1, e2)
      {
          callGeneric(e1, as(e2, "externalVector"))
      })

setMethod("Ops",
          signature(e2="externalVectorWithStorage"),
          function(e1, e2)
      {
          callGeneric(as(e1, "externalVector"), e2)
      })

setMethod("Arith",
          signature(e1="externalVectorWithStorage",
                    e2="externalVectorWithStorage"),
          function(e1, e2)
      {
          ans <- .Call("externalVectorBinaryArith", .Generic,
                       e1, getNativeStorageMethods(e1@storage),
                       e2, getNativeStorageMethods(e2@storage),
                       PACKAGE="externalVector")
          n <- ans@length
          if (length(e1) == n) {
              e1@storage <- ans
              e1
          } else {
              e2@storage <- ans
              e2
          }
      })

setMethod("Arith",
          signature(e1="externalVectorWithStorage",
                    e2="missing"),
          function(e1, e2)
      {
          if (.generic == '+')
              e1
          else {
              type <- e1@storage@type
              if (class(type) == "logical")
                  type <- integer(1)
              ans <- rebind(e1, type = type)
              .Call("externalVectorBinaryArith",
                    e1, getNativeStorageMethods(e1@storage),
                    ans, getNativeStorageMethods(ans@storage),
                    PACKAGE="externalVector")
          }
      })

setMethod("Compare",
          signature(e1="externalVectorWithStorage",
                    e2="externalVectorWithStorage"),
          function(e1, e2)
      {
          ans <- .Call("externalVectorCompare",
                       .Generic,
                       e1, getNativeStorageMethods(e1@storage),
                       e2, getNativeStorageMethods(e2@storage),
                       PACKAGE="externalVector")
          n <- ans@length
          if (length(e1) == n) {
              e1@storage <- ans
              e1
          } else {
              e2@storage <- ans
              e2
          }
      })

setMethod("Complex",
          signature(z="externalVectorWithStorage"),
          function(z)
      {
          ans <- rebind(z,
                        type = if (.Generic == 'Conj' &&
                                   class(x@storage@type) == "complex")
                               complex(1) else numeric(1))
          .Call("externalVectorComplex", .Generic,
                z, getNativeStorageMethods(z@storage),
                ans, getNativeStorageMethods(ans@storage),
                PACKAGE="externalVector")
      })

        setMethod("Summary", signature(x="externalVectorWithStorage"),
                  function(x, ..., na.rm)
              {
                  if (length(list(...)) > 0) {
                      callGeneric(force(x),
                                  callGeneric(..., na.rm=na.rm),
                                  na.rm=na.rm)
                  } else callGeneric(force(x), na.rm=na.rm)
              })

        setMethod("Summary", signature(x="externalVectorWithStorage"),
                  function(x, ..., na.rm)
              {
                  na.rm <- as.logical(na.rm)[1]
                  ## We use .External as we have no idea how many
                  ## elements are there in the ... argument.
                  ans <- .External('externalVectorSummary',
                                   .Generic, na.rm, x, ...,
                                   PACKAGE='externalVector')
                  ## The answer we get above is a list with the
                  ## externalVectorWithStorage arguments evaluated
                  ## and at the end. If such a argument was of
                  ## length zero, it is returned unchanged.
                  ## Other arguments are pushed to the front
                  ## (with those with a class pushed in the front so
                  ## that if there is a Summary method for the
                  ## firt one, it will be used next.)

                  ## Now we create a call that will apply the
                  ## summary method to the elements of the list ANS.
                  mcall <- call('callGeneric')
                  if (length(ans) > 0)
                      mcall[seq(from=2, length=length(ans))] <- ans
                  mcall$na.rm <- na.rm
                  eval(mcall)
              })
        setMethod("Summary",
                  signature(x="indirectExternalMatrix"),
                  function(x, ..., na.rm)
              {
                  callGeneric(force(x), !missing(na.rm) && na.rm)
              })
        setMethod("Summary",
                  signature(x="indirectExternalVector"),
                  function(x, ..., na.rm)
              {
                  callGeneric(force(x), !missing(na.rm) && na.rm)
              })

setMethod("Math",
          signature(x="externalVectorWithStorage"),
          function(x)
      {
          if (.Generic == 'abs') {
              ans <- rebind(x,
                            type = numeric(1))
              return(.Call("externalVectorComplex",
                           .Generic,
                           x, getNativeStorageMethods(x@storage),
                           ans, getNativeStorageMethods(ans@storage),
                           PACKAGE="externalVector"))
          }
          ans <-
              rebind(x,
                     type = if (class(x@storage@type) == "complex") x@storage@type
                     else numeric(1))
          .Call("externalVectorMath", .Generic,
                x, getNativeStorageMethods(x@storage),
                ans, getNativeStorageMethods(ans@storage),
                PACKAGE="externalVector")
      })
