### externalVector class: Vector objects with external storage.

setVirtualMethod('setNames', signature(object='externalVector'),
                 def=function(object, nm) standardGeneric('setNames'))

setVirtualMethod('setDimNames', signature(object='externalVector'),
                 def=function(object, dnm) standardGeneric('setDimNames'))

setReplaceMethod("names",
                 signature(x="externalVector",
                           value="externalVector"),
                 function(x, value)
             {
                 if (class(defaultElement(value)) != "character")
                     value <- asEach(value, "character")
                 if (length(value) > length(x))
                     stop("Names slot must be the same length as the externalVector")
                 if (length(value) < length(x))
                     length(value) <- length(x)
                 setNames(x, value)
             })

setReplaceMethod("names",
                 signature(x="externalVector"),
                 function(x, value)
             {
                 if (!is.character(value))
                     value <- as.character(value)
                 if (length(value) > length(x))
                     stop("Names slot must be the same length as the externalVector")
                 if (length(value) < length(x))
                     length(value) <- length(x)
                 setNames(x, value)
             })

setReplaceMethod("names",
                 signature(x="externalVector", value = "NULL"),
                 function(x, value)
             {
                 if (is.null(names(x)))
                     x
                 else setNames(x, NULL)
             })

setReplaceMethod("dimnames", signature(x="externalVector",
                                       value="NULL"),
                 function(x, value)
             {
                 if (is.null(dim(x)))
                     stop("dimnames applied to non-array")
                 if (is.null(dimnames(x)))
                     x
                 else setDimNames(x, NULL)
             })

setReplaceMethod("dimnames",
                 signature(x="externalVector",
                           value="list"),
                 function(x, value)
             {
                 if (is.null(d <- dim(x)))
                     stop("dimnames applied to non-array")
                 if (length(value) != length(d))
                     stop("length of dimnames must match that of dims")
                 for (i in seq(along=d)) {
                     if (!is.null(value[[i]])) {
                         this <- value[[i]]
                         if (!is.vector(this) && !is(this, "externalVector"))
                             stop("invalid type for dimname (must be a vector)")
                         if (d[i] != length(this) &&
                             length(this) != 0)
                             stop("length of dimnames[",
                                  i, "] not equal to array extent")
                         value[[i]] <-
                             if (length(this) == 0)
                                 NULL
                             else if (is(this,
                                         "externalVector"))
                                 asEach(this, "character")
                             else as.character(this)
                     }
                 }
                 setDimNames(x, value)
             })

setGeneric("getExternalStorage",
               function(x)
               standardGeneric("getExternalStorage"))

setClass("externalVectorWithStorage",
         representation(storage = "externalStorage",
                        Names = "vectorNamesType", # names for the vector
                        Dim = "integerOrNull",     # dimension (may be NULL)
                        DimNames = "listOrNull"),
         contains = "externalVector",
         prototype = list(Names = NULL))

setMethod("initialize",
          signature(.Object="externalVectorWithStorage"),
          function(.Object, ...)
      {
          .Object <- callNextMethod()
          if (class(.Object@storage) == "externalStorage")
              .Object@storage <-
                  new("simpleStorage", type = vector("logical", 1),
                      length = as.integer(0))
          .Object[] <- defaultElement(.Object)
          .Object
      })

setMethod("getExternalStorage",
          signature(x="externalVectorWithStorage"),
          function (x)
      {
          x@storage
      })

setIs("externalVectorWithStorage", "vectorNamesType")

setMethod("getNativeStorageMethods", "externalVectorWithStorage",
          function(x)
      {
          getNativeStorageMethods(x@storage)
      })

setMethod("show",
          signature(object = "externalVector"),
          function(object)
      {
          objDim <- dim(object)
          cat("External",
              if (is.null(objDim)) "vector"
              else if (length(objDim) == 2) "matrix"
              else "array",
              "object with\n")
          cat("  Class:", class(object), "\n")
          cat("  Storage:", class(getExternalStorage(object)), "\n")
          cat("  Type:", class(defaultElement(object)), "\n")
          if (is.null(objDim) || length(objDim) == 1) {
              n <- length(object)
              cat("  Length:", n, "\n")
              if (n > 0) {
                  if (n > 5) {
                      object <- as.Rvector(force(object[c(1:4, n)]))
                      nm <- names(object)
                      object <-
                          if (is.character(object)) {
                              paste('"', object, '"', sep = '')
                          } else format(object)
                      object <- c(object[1:4], "...", object[5])
                      names(object) <-
                          if (is.null(nm))
                              c(1:4, '', n)
                          else c(nm[1:4], '', nm[5])
                  } else {
                      object <- as.Rvector(force(object))
                      object <-
                          if (is.character(object))
                              paste('"', object, '"', sep = '')
                          else format(object)
                  }
              }
          } else {
              cat("  Dim:", objDim, "\n")
              if (length(object) > 0) {
                  if (length(objDim) == 2) {
                      i <- if (objDim[1] > 5) c(1:4, objDim[1]) else seq(length=objDim[1])
                      j <- if (objDim[2] > 5) c(1:4, objDim[2]) else seq(length=objDim[2])
                      object <- force(object[i, j, drop = FALSE])[,,drop=FALSE]
                      dnm <- dimnames(object)
                      if (is.null(dnm))
                          dnm <- list(NULL, NULL)
                      if (is.null(dnm[[1]])) {
                          dnm[[1]] <- format(paste('[', i, ',]', sep=''),
                                             justify='right')
                      }
                      if (is.null(dnm[[2]])) {
                          dnm[[2]] <- paste('[,', j, ']', sep='')
                      }
                      if (length(i) != objDim[1]) {
                          mid <- ceiling(unlist(lapply(object[4, ], nchar))/2)
                          dots <- character(mid)
                          dots[] <- ' '
                          dots[mid] <- '.'
                          dots <- paste(dots, collapse = '')
                          object <- rbind(object[1:4, , drop = FALSE],
                                          matrix(rep(dots, 3*ncol(object)),
                                                 nrow=3),
                                          object[5, ])
                          dnm[[1]] <- c(dnm[[1]][1:4], rep('', 3), dnm[[1]][5])
                      }
                      if (length(j) != objDim[2]) {
                          dnm[[2]] <- c(dnm[[2]][1:4], '', dnm[[2]][5])
                          object <- format(rbind(dnm[[2]],
                                                 cbind(object[, 1:4, drop = FALSE],
                                                       if (length(i) != objDim[1]) {
                                                           rep('...', 8)
                                                       } else {
                                                           rep('...', objDim[1])
                                                       },
                                                       object[, 5])),
                                           justify = 'right')
                          dnm[[2]] <- object[1, ]
                          object <- object[-1, , drop = FALSE]
                      }
                      dimnames(object) <- dnm
                  } else if (length(objDim) > 2) {
                      object <- NULL
                  }
              }
          }
          if (length(object) > 0)
              print(object, quote=FALSE)
      })

setMethod("dim", "externalVectorWithStorage",
          function(x) x@Dim)

setReplaceMethod("dim", signature(x="externalVectorWithStorage",
                                  value="NULL"),
                 function(x, value)
             {
                 if (!is.null(x@Dim)) {
                     if (!is.null(x@DimNames))
                         x@DimNames <- NULL
                     x@Dim <- NULL
                 }
                 names(x) <- NULL
                 x
             })

setReplaceMethod("dim", signature(x="externalVectorWithStorage"),
                 function(x, value)
             {
                 value <- as.integer(value)
                 if (length(value) == 0)
                     stop("dim: Invalid dimension vector")
                 if (length(x) != prod(value))
                     stop("dim<- length of dims do not match the length of object")
                 x@Dim <- value
                 if (!is.null(x@DimNames))
                     x@DimNames <- NULL
                 names(x) <- NULL
                 x
             })

setMethod("dimnames", "externalVectorWithStorage",
          function(x) x@DimNames)

rebind <-
    function(x, length.x = length(x), type = x@storage@type,
             names.x = x@Names, dim.x = x@Dim,
             dimnames.x = x@DimNames, ..., arrayOnly = FALSE)
{
    if (!is(x, "externalVectorWithStorage"))
        stop("x must be from class externalVectorWithStorage")
    if (missing(type))
        typeName <- class(x@storage@type)
    else {
        typeName <- if (is.character(type))
            type
        else class(type)
    }
    if (typeName %in% c("logical", "numeric", "double", "integer", "complex",
                        "character"))
        type <- vector(typeName, 1)
    else if (is.character(type))
        type <- new(type)
    length.x <- as.integer(length.x)
    if (as.logical(arrayOnly)) {
        new(class(x), storage = new(class(x@storage), type=type,
                      length=length.x))
    } else {
        new(class(x), storage = new(class(x@storage), type=type,
                      length=length.x),
            Names = names.x, Dim = dim.x, DimNames = dimnames.x, ...)
    }
}

setGeneric("defaultElement",
               function(x)
               standardGeneric("defaultElement"))

setMethod("defaultElement", "externalVector",
          function(x) getExternalStorage(x)@type)

setMethod("defaultElement", "vector",
          function(x) vector(class(x), 1))

setMethod("length", "externalVectorWithStorage",
          function(x)
          x@storage@length)

setReplaceMethod("length", signature(x = "externalVectorWithStorage"),
                 function(x, value)
             {
                 value <- as.integer(value)
                 if (value != length(x)) {
                     x <- .Call("externalVector_setLength", x,
                                getNativeStorageMethods(x),
                                value,
                                PACKAGE="externalVector")
                 }
                 x
             })

setGeneric("asEach",
               function(x, type, arrayOnly=FALSE)
               standardGeneric("asEach"))

setMethod("asEach",
          signature(type="character"),
          function (x, type, arrayOnly)
      {
          if (arrayOnly)
              as.vector(as.Rvector(x), type)
          else {
              saveAttribs <- nonSlotAttributes(x)
              saveAttribs$dim <- dim(x)
              saveAttribs$dimnames <- dim(x)
              saveAttribs$names <- dim(x)
              x <- as.vector(as.Rvector(x), type)
              attributes(x) <- saveAttribs
              x
          }
      })

setMethod("asEach",
          signature(x="externalVectorWithStorage", type = "character"),
          function(x, type, arrayOnly)
      {
          if (defaultElement(x) == type)
              x
          else {
              y <- rebind(x, type = type, arrayOnly=arrayOnly)
              .Call("externalVector_asEach", x,
                    getNativeStorageMethods(x@storage),
                    y, getNativeStorageMethods(y@storage),
                    PACKAGE = "externalVector")
          }
      })

setMethod("names", "externalVectorWithStorage",
          function(x) x@Names)

setReplaceMethod("names",
                 signature(x="externalVectorWithStorage", value = "NULL"),
                 function(x, value)
             {
                 if (!is.null(x@Names))
                     x@Names <- NULL
                 x
             })

setMethod("setNames", signature(object="externalVectorWithStorage"),
          function(object, nm)
      {
          object@Names <- nm
          object
      })

setMethod("setDimNames", signature(object="externalVectorWithStorage"),
          function(object, dnm)
      {
          object@DimNames <- dnm
          object
      })

setMethod("match", signature(x="externalVectorWithStorage"),
          function(x, table, nomatch, incomparables)
      {
          if (missing(nomatch))
              nomatch <- as.integer(NA)
          else nomatch <- as.integer(nomatch)

          n <- length(x)

          ## the class of answer will have the same subclass of
          ## externalVector but with type integer
          ans <- rebind(x, type = integer(1))

          ## split the problem into pieces of length 65536
          nsplits <- n / 65536
          nleft <- n %% 65536
          indx <- 1:65536

          ## match each piece of x
          for (i in seq(length = nsplits)) {
              ans[indx] <- match(x[indx], table, nomatch)
              indx[] <- indx + 65536
          }
          if (nleft > 0) {
              length(indx) <- nleft
              ans[indx] <- match(x[indx], table, nomatch)
          }

          ans
      })

setMethod("match", signature(table="externalVectorWithStorage"),
          function(x, table, nomatch, incomparables)
      {
          if (length(x) == 0)
              return (integer(0))

          n <- length(table)
          ans <- integer(length(x))

          ## This can be done more efficiently (possibly in C)

          ## split the problem in to pieces of length 65536
          nsplits <- n / 65536
          nleft <- n %% 65536
          indx <- 1:65536
          unmatched <- ans

          ## match each piece of x
          for (i in seq(length = nsplits)) {
              ans[unmatched] <- ans + match(x[unmatched], table[indx],
                                            nomatch = 1-indx[1]) + indx[1]-1
              unmatched <- ans == 0
              if (!any(unmatched)) {
                  nleft <- 0
                  break
              }
              indx[] <- indx + 65536
          }
          if (nleft > 0) {
              length(indx) <- nleft
              ans[unmatched] <- ans + match(x[unmatched], table[indx],
                                            nomatch = 1-indx[1]) + indx[1]-1
              unmatched <- ans == 0
          }
          ans[unmatched] <- if (missing(nomatch))
              as.integer(NA)
          else as.integer(nomatch)
          ans
      })

setMethod("[",
          signature(x="externalVectorWithStorage"),
          function(x, i, j, ..., drop)
      {
          if (missing(drop)) {
              drop <- FALSE
              nsubs <- nargs() - 1
          } else {
              nsubs <- nargs() - 2
          }
          if (nsubs == 0)
              nsubs <- 1
          if (nsubs == 1 && missing(i))
              x
          else {
              subs <-
                  if (nsubs == 1) {
                      alist(i=)
                  } else if (nsubs == 2) {
                      alist(i=, j=)
                  } else {
                      alist(i=, j=, ...)
                  }
              if (!missing(i))
                  subs['i'] <- list(i)
              if (!missing(j))
                  subs['j'] <- list(j)
              .Call("externalSubset", x,
                    getNativeStorageMethods(x@storage),
                    getNativeStorageMethods(names(x)),
                    SubscriptList(subs, length(x), dim(x),
                                  names(x), dimnames(x), FALSE), drop,
                    PACKAGE="externalVector")
          }
      })

setReplaceMethod("[", signature(x="externalVectorWithStorage",
                                value="externalVectorWithStorage"),
                 function(x, i, j, ..., value)
             {
                 nsubs <- nargs() - 2
                 if (nsubs == 0)
                     nsubs <- 1
                 subs <-
                     if (nsubs == 1) {
                         alist(i=)
                     } else if (nsubs == 2) {
                         alist(i=, j=)
                     } else {
                         alist(i=, j=, ...)
                     }
                 if (!missing(i))
                     subs$i <- i
                 if (!missing(j))
                     subs$j <- j
                 subs <- SubscriptList(subs, length(x), dim(x),
                                       names(x), dimnames(x), TRUE)
                 .Call("externalSubassign", x,
                       getNativeStorageMethods(x@storage),
                       subs, attr(subs, 'newnames'),
                       attr(subs, 'newlength'), value,
                       getNativeStorageMethods(value@storage),
                       PACKAGE="externalVector")
             })

setReplaceMethod("[", "externalVectorWithStorage",
                 function(x, i, j, ..., value)
             {
                 nsubs <- nargs() - 2
                 if (nsubs == 0)
                     nsubs <- 1
                 subs <-
                     if (nsubs == 1) {
                         alist(i=)
                     } else if (nsubs == 2) {
                         alist(i=, j=)
                     } else {
                         alist(i=, j=, ...)
                     }
                 if (!missing(i))
                     subs$i <- i
                 if (!missing(j))
                     subs$j <- j
                 subs <- SubscriptList(subs, length(x), dim(x),
                                       names(x), dimnames(x), TRUE)
                 .Call("externalSubassign", x,
                       getNativeStorageMethods(x@storage),
                       subs, attr(subs, 'newnames'),
                       attr(subs, 'newlength'), value, NULL,
                       PACKAGE="externalVector")
             })

setMethod("is.finite",
          signature(x = "externalVectorWithStorage"),
          function (x)
          is.finite(as.Rvector(x)))

setMethod("is.infinite",
          signature(x = "externalVectorWithStorage"),
          function (x) is.infinite(as.Rvector(x)))

setMethod("is.na",
          signature(x = "externalVectorWithStorage"),
          function (x) is.na(as.Rvector(x)))

setMethod("is.nan",
          signature(x = "externalVectorWithStorage"),
          function (x) is.nan(as.Rvector(x)))
