setClass("vectorIndex")

setVirtualMethod("Arith",
                 signature(e1 = "vectorIndex"))

setVirtualMethod("Arith",
                 signature(e2 = "vectorIndex"))

setClass("NAIndex",
         representation(Length="integer", Names="vectorNamesType"),
         contains = c("vectorIndex", "logical"),
         prototype=structure(NA, length=as.integer(1),
                             Names = NULL))

setMethod("initialize",
          signature(.Object = "NAIndex"),
          function (.Object, Length = 0, ...)
      {
          .Object <- .Call("initNAIndex", Length,
                           PACKAGE="externalVector")
          if (!missing(Names)) {
              Names <- asEach(Names, "character")
              if (length(.Object) < length(Names))
                  stop("Names slot must be the same length as the vector")
              else if (length(.Object) > length(Names))
                  length(Names) <- length(.Object)
              .Object@Names <- Names
          }
          .Object
      })

setClass("allIndex",
         contains=c("vectorIndex", "logical"),
         prototype=TRUE)

setMethod("initialize",
          signature(.Object = "allIndex"),
          function (.Object, ...) .Call("initAllIndex",
                                        PACKAGE="externalVector"))

setClass("noneIndex", contains=c("vectorIndex", "integer"),
         prototype=integer(0))

setMethod("initialize",
          signature(.Object = "noneIndex"),
          function (.Object, ...) .Call("initNoneIndex",
                                        PACKAGE="externalVector"))

setClass("sliceIndex",
         representation(content="numeric", Names="vectorNamesType"),
         contains="vectorIndex",
         prototype=list(content=as.numeric(rep(1, 3))))

setClass("positiveIndex",
         representation(rangeIndex="integer",
                        noNA="logical"),
         contains = c("vectorIndex", "integer"),
         prototype = structure(integer(0),
                               rangeIndex=as.integer(c(NA, NA)),
                               noNA=NA))

setMethod("initialize",
          signature(.Object = "positiveIndex"),
          function (.Object, ...) {
              .External('initPositiveIndex', ...,
                    PACKAGE='externalVector')
          })

setMethod("names",
          signature(x="NAIndex"),
          function (x) x@Names)

setMethod("names",
          signature(x="sliceIndex"),
          function (x) x@Names)

setReplaceMethod("names",
                 signature(x = "NAIndex"),
                 function (x, value)
             {
                 x@Names <- as.character(value)
                 x
             })

setReplaceMethod("names",
                 signature(x = "NAIndex"),
                 function (x, value)
             {
                 if (!is.character(value))
                     value <- as.character(value)
                 if (length(value) > length(x))
                     stop("Names slot must be the same length as the externalVector")
                 if (length(value) < length(x))
                     length(value) <- length(x)
                 x@Names <- value
                 x
             })

setReplaceMethod("names",
                 signature(x = "NAIndex", value = "externalVector"),
                 function (x, value)
             {
                 if (class(defaultElement(value)) != "character")
                     value <- asEach(value, "character")
                 if (length(value) > length(x))
                     stop("Names slot must be the same length as the externalVector")
                 if (length(value) < length(x))
                     length(value) <- length(x)
                 x@Names <- x
                 x
             })

setReplaceMethod("names",
                 signature(x = "sliceIndex"),
                 function (x, value)
             {
                 x@Names <- as.character(value)
                 x
             })

setReplaceMethod("names",
                 signature(x = "sliceIndex"),
                 function (x, value)
             {
                 if (!is.character(value))
                     value <- as.character(value)
                 if (length(value) > length(x))
                     stop("Names slot must be the same length as the externalVector")
                 if (length(value) < length(x))
                     length(value) <- length(x)
                 x@Names <- value
                 x
             })

setReplaceMethod("names",
                 signature(x = "sliceIndex", value = "externalVector"),
                 function (x, value)
             {
                 if (class(defaultElement(value)) != "character")
                     value <- asEach(value, "character")
                 if (length(value) > length(x))
                     stop("Names slot must be the same length as the externalVector")
                 if (length(value) < length(x))
                     length(value) <- length(x)
                 x@Names <- x
                 x
             })

setMethod("anyNA",
          signature(x="allIndex"),
          function (x) FALSE)

setMethod("anyNA",
          signature(x="noneIndex"),
          function (x) FALSE)

setMethod("anyNA",
          signature(x="NAIndex"),
          function (x) TRUE)

setMethod("anyNA",
          signature(x="sliceIndex"),
          function (x) FALSE)

setMethod("anyNA",
          signature(x="positiveIndex"),
          function (x)
          if (is.na(x@noNA)) {
              any(is.na(x))
          } else !x@noNA)

setMethod("allNA",
          signature(x="allIndex"),
          function (x) FALSE)

setMethod("allNA",
          signature(x="noneIndex"),
          function (x) FALSE)

setMethod("allNA",
          signature(x="NAIndex"),
          function (x) TRUE)

setMethod("allNA",
          signature(x="sliceIndex"),
          function (x) FALSE)

setMethod("allNA",
          signature(x="positiveIndex"),
          function (x)
          (is.na(x@noNA) || !x@noNA) && all(is.na(x)))

setMethod("length",
          signature(x = "NAIndex"),
          function (x) x@Length)

setReplaceMethod("length",
                 signature(x = "NAIndex"),
                 function (x, value)
             {
                 n <- as.integer(value)
                 if (is.na(value))
                     stop("missing value for length")
                 if (length(value) > 1)
                     stop("invalid second argument")
                 if (length(value) != 1)
                     stop("invalid second argument")
                 if (value < 0)
                     stop("negative length vectors are not allowed")
                 if (value == 0)
                     new("noneIndex")
                 else {
                     x@Length <- n
                     if (!is.null(names(x)))
                         length(x@Names) <- n
                     x
                 }
             })

setMethod("[",
          signature(x = "NAIndex"),
          function (x, i, j, ..., drop)
      {
          length(x) <- length(i)
          x
      })

setMethod("[",
          signature(x = "NAIndex", i = "noneIndex"),
          function (x, i, j, ..., drop)
      {
          i
      })

setMethod("[",
          signature(x = "NAIndex", i = "allIndex"),
          function (x, i, j, ..., drop)
      {
          x
      })

setReplaceMethod("[",
                 signature(x = "NAIndex"),
                 function (x, i, j, ..., value)
             {
                 stop("object is not subassignable")
             })

setReplaceMethod("[[",
                 signature(x = "NAIndex"),
                 function (x, i, j, ..., value)
             {
                 stop("object is not subassignable")
             })

setMethod("length",
          signature(x = "allIndex"),
          function (x)
      {
          as.integer(Inf)
      })

setReplaceMethod("length",
                 signature(x = "allIndex"),
                 function (x, value)
             {
                 stop("can not change length of object")
             })

setMethod("[",
          signature(x = "allIndex", i = "ANY",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop)
      {
          i
      })

setMethod("[",
          signature(x = "allIndex", i = "missing",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop)
      {
          x
      })

setReplaceMethod("[",
                 signature(x = "allIndex"),
                 function (x, i, j, ..., value)
             {
                 stop("object is not subassignable")
             })

setReplaceMethod("[[",
                 signature(x = "allIndex"),
                 function (x, i, j, ..., value)
             {
                 stop("object is not subassignable")
             })

setMethod("length",
          signature("noneIndex"),
          function (x)
      {
          as.integer(0)
      })

setMethod("[",
          signature(x = "noneIndex", i = "missing",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop)
      {
          x
      })

setReplaceMethod("length",
                 signature(x = "noneIndex"),
                 function (x, value)
             {
                 stop("can not change length of object")
             })

setReplaceMethod("[",
                 signature(x = "noneIndex"),
                 function (x, i, j, ..., value)
             {
                 stop("object is not subassignable")
             })

setReplaceMethod("[[",
                 signature(x = "noneIndex"),
                 function (x, i, j, ..., value)
             {
                 stop("object is not subassignable")
             })

setMethod("[",
          signature(x = "noneIndex", i = "ANY",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop)
      {
          new("NAIndex", length=length(i))
      })

setMethod("[",
          signature(x = "noneIndex", i = "logical",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop)
      {
          new("NAIndex", length=sum(is.na(i), i, na.rm=TRUE))
      })

setMethod("length",
          signature(x = "sliceIndex"),
          function (x)
      {
          as.integer(x@content[2])
      })

# setMethod("rev",
#           signature(x="sliceIndex"),
#           function (x)
#       {
#           if (length(x) > 1) {
#               e1@content[3] = -e1@content[3]
#               e1@content[1] = e1@content[1]-e1@content[3]*(e1@content[2]-1)
#           }
#           x
#       })

setReplaceMethod("length",
                 signature(x="sliceIndex"),
                 function(x, value)
             {
                 if (is.nan(value) || value == -Inf)
                     stop('length value invalid')
                 if (value < 1.0)
                     return(new("noneIndex"))
                 value <- trunc(as.numeric(value))
                 if (value != x@content[1]) {
                     x@content[1] <- value
                     if (!is.null(names(x)))
                         length(x@Names) <- value
                 }
                 x
             })

setMethod("[",
          signature(x = "sliceIndex", i = "allIndex",
                    j = "missing"),
          function (x, i, j, ..., drop)
      {
          x
      })

setMethod("[",
          signature(x = "sliceIndex", i = "noneIndex",
                    j = "missing"),
          function (x, i, j, ..., drop)
      {
          i
      })

setMethod("[",
          signature(x = "sliceIndex", i = "sliceIndex",
                    j = "missing"),
          function (x, i, j, ..., drop)
      {
          contx <- x@content
          conti <- i@content
          start <- conti[1]
          len <- conti[2]
          end <- conti[1]+conti[3]*(len-1)
          if (start >= 0 && end < contx[2]) {
              conti[1] <- contx[1]+start*contx[3]
              conti[3] <- contx[3]*conti[3]
              x@content <- conti
              if (!is.null(names(x)))
                  names(x) <- Subset(names(x), i)
              x
          } else as(x, "positiveIndex")[i]
      })

setMethod("[",
          signature(x = "sliceIndex", i = "positiveIndex",
                    j = "missing"),
          function (x, i, j, ..., drop)
      {
          nm <- x@Names
          contx <- x@content
          start <- contx[1]
          len <- contx[2]
          stride <- contx[3]

          noNA <- i@noNA
          if (is.na(noNA)) {
              na <- is.na(i)
              noNA <- any(na)
              if (noNA)
                  na <- logical(0)
          } else if (noNA)
              na <- logical(0)
          else na <- is.na(i)

          range.i <- i@rangeIndex
          if (anyNA(range.i))
              range.i <- as.integer(range(i, na.rm=T))

          if (range.i[1] < 0 || range.i[2] >= len) {
              if (noNA) {
                  na <- i < 0 | i >= len
                  if (!any(na))
                      na <- logical(0)
              } else na[!na] <- i[!na] < 0 | i[!na] >= len
          }
          if (length(na) == 0) {
              ans <- start+i*stride
          } else {
              ans <- start+i*stride
              ans[na] <- as.integer(NA)
              ans@noNA <- FALSE
          }
          if (!is.null(nm))
              names(ans) <- Subset(nm, i)
          ans
      })

setMethod("[",
          signature(x = "sliceIndex",
                    j = "missing"),
          function (x, i, j, ..., drop)
      {
          as(x, "positiveIndex")[i]
      })

setMethod("[",
          signature(x = "positiveIndex", i = "sliceIndex",
                    j = "missing"),
          function (x, i, j, ..., drop)
      {
          conti <- as.integer(i@content)
          start <- conti[1]
          len <- conti[2]
          stride <- conti[3]
          if (len > 0) {
              i <- start+1+(0:(len-1))*stride
              as(as.integer(x)[i], "positiveIndex")
          } else new("positiveIndex", integer(0))
      })

setMethod("[",
          signature(x = "positiveIndex", i = "positiveIndex",
                    j = "missing"),
          function (x, i, j, ..., drop)
      {
          nm <- names(x)
          if (!is.null(nm))
              nm <- Subset(nm, i)
          i <- unclass(i+1)
          x <- new("positiveIndex", unclass(x)[i])
          names(x) <- nm
          x
      })

setMethod("[",
          signature(x = "positiveIndex",
                    j = "missing"),
          function (x, i, j, ..., drop)
      {
          as(Subset(as.integer(x), i), "positiveIndex")
      })

makeSlice <-
    function(start=1, length.slice=1, stride=1)
{
    .Call("makeSlice", start, length.slice, stride,
          PACKAGE="externalVector")
}

setAs("integer", "positiveIndex",
      function(from, to) {
          if (any(from < 0, na.rm=TRUE))
              stop("negative integers can not be coerced to positiveIndex")
          from <- from[from != 0]
          nm <- names(from)
          names(from) <- NULL
          ans <- new("positiveIndex", as.integer(from-1))
          names(ans) <- nm
          ans
      })

setGeneric("as.integer")

setMethod("as.integer",
          signature(x="positiveIndex"),
          function (x, ...)
      {
          nm <-names(x)
          x <- as.integer(as.vector(x)+1)
          names(x) <- nm
          x
      })

setAs("numeric", "positiveIndex",
      function(from, to) {
          as(as.integer(from), "positiveIndex")
      })

setAs("sliceIndex", "positiveIndex",
      function(from, to) {
          cont <- as.integer(from@content)
          n <- cont[2]
          ans <- if (n > 0)
              new("positiveIndex",
                  cont[1]+(0:(n-1))*cont[3])
          else new("positiveIndex", integer(0))
          names(ans) <- names(from)
          ans
      })

setMethod("Arith",
          signature(e1 = "noneIndex", e2 = "numeric"),
          function (e1, e2) e1)

setMethod("Arith",
          signature(e1 = "numeric", e2 = "noneIndex"),
          function (e1, e2) e2)

setMethod("Arith",
          signature(e1 = "NAIndex", e2 = "numeric"),
          function (e1, e2) e1)

setMethod("Arith",
          signature(e1 = "numeric", e2 = "NAIndex"),
          function (e1, e2) e2)

setMethod("Arith",
          signature(e1 = "allIndex", e2 = "numeric"),
          function (e1, e2) e1)

setMethod("Arith",
          signature(e1 = "numeric", e2 = "allIndex"),
          function (e1, e2) e2)

setMethod("Arith",
          signature(e1 = "sliceIndex", e2 = "numeric"),
          function (e1, e2)
      {
          if (length(e2) != 1 || !(.Generic %in% c('+', '*')))
              stop('unimplemented operation')
          if (is.na(as.integer(e2))) {
              new("NAIndex", length=as.integer(e1@content[2]),
                  Names = names(e1))
          } else {
              if (.Generic == '+')
                  e1@content[1] <- e1@content[1]+e2
              else e1@content[c(1, 3)] <- e1@content[c(1, 3)]*e2
              if (e1@content[1] < 0 ||
                  e1@content[1]+(e1@content[2]-1)*e1@content[3] < 0)
                  stop("negative index not allowed")
              e1
          }
      })

setMethod("Arith",
          signature(e1 = "numeric", e2 = "sliceIndex"),
          function (e1, e2)
          callGeneric(e2, e1))

setMethod("Arith",
          signature(e1 = "positiveIndex", e2 = "numeric"),
          function (e1, e2)
      {
          if (length(e2) != 1 || !(.Generic %in% c('+', '*')))
              stop('unimplemented operation')
          if (is.na(as.integer(e2)))
              new("NAIndex", length=length(e1), Names = names(e1))
          else {
              e1 <- e1@.Data
              e2 <- as.integer(e2)
              e1 <- if (.Generic == '+')
                  e1+e2
              else e1*e2
              new("positiveIndex", e1)
          }
      })

setMethod("Arith",
          signature(e1 = "numeric", e2 = "positiveIndex"),
          function (e1, e2)
          callGeneric(e2, e1))

SubscriptList <-
    function(subs, length.object, dim.object=NULL,
             names.object=NULL, dimnames.object=NULL,
             assign.operation=FALSE)
{
    .Call("SubscriptList", subs, length.object, dim.object,
          names.object, dimnames.object, assign.operation,
          PACKAGE='externalVector')
}

setGeneric("Subset",
               function(x, i) standardGeneric("Subset"))

setMethod("Subset",
          signature("ANY"),
          function (x, i)
      {
          x[i]
      })

setMethod("Subset",
          signature(x = "ANY", i = "allIndex"),
          function (x, i) x)

setMethod("Subset",
          signature(x = "ANY", i = "noneIndex"),
          function (x, i) x[NULL])

setMethod("Subset",
          signature(x = "ANY", i = "NAIndex"),
          function (x, i)
      {
          na <- x[length(x)+1]
          length(x) <- length(i)
          x[] <- na
          x
      })

setMethod("Subset",
          signature(x = "ANY", i = "positiveIndex"),
          function (x, i)
      {
          zero <- i == 0
          i[zero] <- NA
          i <- x[-1][i]
          i[zero] <- x[1]
          i
      })

setMethod("Subset",
          signature(x = "ANY", i = "sliceIndex"),
          function (x, i)
      {
          cont <- as.integer(i@content)
          start <- cont[1]
          n <- cont[2]
          stride <- cont[3]
          x[start+1+(0:(n-1))*stride]
      })

setGeneric("Subset2",
               function(x, i, j) standardGeneric("Subset2"))

setMethod("Subset2",
          signature("ANY"),
          function (x, i, j)
      {
          x[i, j, drop = FALSE]
      })

setMethod("Subset2",
          signature(x = "ANY", i = "allIndex", j = "allIndex"),
          function (x, i, j) x)

setMethod("Subset2",
          signature(x = "ANY", i = "allIndex"),
          function (x, i, j) x[, j, drop=FALSE])

setMethod("Subset2",
          signature(x = "ANY", j = "allIndex"),
          function (x, i, j) x[i, , drop=FALSE])

setMethod("Subset2",
          signature(x = "ANY", i = "noneIndex"),
          function (x, i, j) Subset(x, NULL, j))

setMethod("Subset2",
          signature(x = "ANY", j = "noneIndex"),
          function (x, i, j) Subset(x, i, NULL))

setMethod("Subset2",
          signature(x = "ANY", i = "noneIndex", j = "noneIndex"),
          function (x, i, j) x[NULL, NULL, drop = FALSE])

setMethod("Subset2",
          signature(x = "ANY", i = "NAIndex", j = "NAIndex"),
          function (x, i, j)
      {
          na <- x[length(x)+1]
          length(x) <- length(i)*length(j)
          dim(x) <- c(length(i), length(j))
          x[,] <- na
          x
      })

setMethod("Subset2",
          signature(x = "ANY", i = "NAIndex"),
          function (x, i, j)
      {
          na <- x[length(x)+1]
          length(x) <- length(i)*length(j)
          dim(x) <- c(length(i), length(j))
          x[,] <- na
          x
      })

setMethod("Subset2",
          signature(x = "ANY", j = "NAIndex"),
          function (x, i, j)
      {
          na <- x[length(x)+1]
          length(x) <- length(i)*length(j)
          dim(x) <- c(length(i), length(j))
          x[,] <- na
          x
      })

setMethod("Subset2",
          signature(x = "ANY", i = "NAIndex", j = "allIndex"),
          function (x, i, j)
      {
          na <- x[length(x)+1]
          n <- ncol(x)
          length(x) <- length(i)*n
          dim(x) <- c(length(i), n)
          x[,] <- na
          x
      })

setMethod("Subset2",
          signature(x = "ANY", i = "allIndex", j = "NAIndex"),
          function (x, i, j)
      {
          na <- x[length(x)+1]
          m <- nrow(x)
          length(x) <- m*length(j)
          dim(x) <- c(m, length(j))
          x[,] <- na
          x
      })

setMethod("Subset2",
          signature(x = "ANY", i = "positiveIndex", j = "positiveIndex"),
          function (x, i, j)
      {
          x[unclass(i+1), unclass(j+1), drop = FALSE]
      })

setMethod("Subset2",
          signature(x = "ANY", i = "positiveIndex"),
          function (x, i, j)
      {
          Subset2(x, unclass(i+1), j)
      })

setMethod("Subset2",
          signature(x = "ANY", j = "positiveIndex"),
          function (x, i, j)
      {
          Subset2(x, i, unclass(j+1))
      })

setMethod("Subset2",
          signature(x = "ANY", i = "sliceIndex", j = "sliceIndex"),
          function (x, i, j)
      {
          cont <- as.integer(i@content)
          start1 <- cont[1]
          n1 <- cont[2]
          stride1 <- cont[3]
          cont <- as.integer(j@content)
          start2 <- cont[1]
          n2 <- cont[2]
          stride2 <- cont[3]
          x[start1+1+(0:(n1-1))*stride1,
            start2+1+(0:(n2-1))*stride2, drop=FALSE]
      })

setMethod("Subset2",
          signature(x = "ANY", i = "sliceIndex"),
          function (x, i, j)
      {
          cont <- as.integer(i@content)
          start <- cont[1]
          n <- cont[2]
          stride <- cont[3]
          Subset2(x, start+1+(0:(n-1))*stride, j)
      })

setMethod("Subset2",
          signature(x = "ANY", j = "sliceIndex"),
          function (x, i, j)
      {
          cont <- as.integer(j@content)
          start <- cont[1]
          n <- cont[2]
          stride <- cont[3]
          Subset2(x, i, start+1+(0:(n-1))*stride)
      })
