setClass("indirectExternalVector",
         representation(actual="externalVector",
                        index = "vectorIndex"),
         contains = "externalVector")

if (!isClass("exprMatrix"))
    setClassUnion("exprMatrix", c("matrix"))

## Class union for the exprs and se.exprs slots of exprSet
if (!isClass("exprMatrix"))
    setClassUnion("exprMatrix", c("matrix"))

setClass("indirectExternalMatrix",
         representation(actual="externalVector",
                        index = "matrixIndex"),
         contains = c("externalVector"))

setIs("indirectExternalMatrix", "exprMatrix")

setMethod("getExternalStorage",
          signature(x="indirectExternalVector"),
          function (x) getExternalStorage(x@actual))

setGeneric("force")

setMethod("names", "indirectExternalVector",
          function(x) names(x@index))

setMethod("length",
          signature(x = "indirectExternalVector"),
          function (x)
      {
          length(x@index)
      })

setMethod("force",
          signature(x="indirectExternalVector"),
          function (x)
      {
          ans <- force(x@actual)
          if (is(x@index, "sliceIndex") && x@index@content == c(0,
                                                   length(ans), 1.0)) {
              dim(ans) <- NULL
          } else {
              ans <- Subset(ans, x@index)
          }
          names(ans) <- names(x)
          ans
      })

setReplaceMethod("names",
                 signature(x="indirectExternalVector", value = "NULL"),
                 function(x, value)
             {
                 names(x@index) <- NULL
                 x
             })

setMethod("setNames",
          signature(object="indirectExternalVector"),
          function(object, nm)
      {
          names(object@index) <- nm
          object
      })

setReplaceMethod("length",
                 signature(x = "indirectExternalVector"),
                 function (x, value)
             {
                 stop("cannot modify length of indirect external vector")
             })

setMethod("dim",
          signature(x = "indirectExternalVector"),
          function (x) NULL)

setReplaceMethod("dim", signature(x="indirectExternalVector"),
                 function(x, value)
             {
                 value <- as.integer(value)
                 if (length(value) == 0)
                     stop("dim: Invalid dimension vector")
                 if (length(x@actual) != prod(value))
                     stop("dim<- length of dims do not match the length of original object")
                 indx <- x@index
                 dim(indx) <- value
                 indx@origdim <- value
                 new("indirectExternalMatrix",
                     actual=x@actual,
                     index=indx)
             })

setMethod("dimnames",
          signature(x = "indirectExternalVector"),
          function (x) NULL)

setReplaceMethod("dimnames",
                 signature(x = "indirectExternalVector"),
                 function (x, value)
                 stop("dimnames applied to non-array"))

setMethod("[",
          signature(x = "indirectExternalVector", i = "missing",
                    j = "missing"),
          function (x, i, j, ..., drop) x)

setMethod("[",
          signature(x = "indirectExternalVector", j = "missing"),
          function (x, i, j, ..., drop)
      {
          i <- x@index[i]
          if (anyNA(i))
              stop('subscript out of bounds')
          new("indirectExternalVector",
              actual=x@actual,
              index=i)
      })

setReplaceMethod("[",
                 signature(x = "indirectExternalVector", i = "missing",
                           j = "missing"),
                 function (x, i, j, ..., value)
             {
                 x@actual[x@index] <- value
                 x
             })

setReplaceMethod("[",
                 signature(x = "indirectExternalVector", j = "missing"),
                 function (x, i, j, ..., value)
             {
                 i <- x@index[i]
                 if (anyNA(i))
                     stop('subscript out of bounds')
                 ans <- new("indirectExternalVector",
                            actual=x@actual,
                            index=i)
                 ans@actual[i] <- value
                 ans
             })

setMethod("getExternalStorage",
          signature(x="indirectExternalMatrix"),
          function (x) getExternalStorage(x@actual))

setMethod("force",
          signature(x="indirectExternalMatrix"),
          function (x)
      {
          ans <- force(x@actual)
          dim(ans) <- x@index@origdim
          if (!is(x@index@row, "sliceIndex") ||
              !is(x@index@col, "sliceIndex") ||
              any(x@index@row@content != c(0, nrow(ans), 1.0)) ||
              any(x@index@col@content != c(0, ncol(ans), 1.0))) {
              ans <- Subset(ans, x@index)
          }
          dim(ans) <- dim(x)
          dimnames(ans) <- dimnames(x)
          names(ans) <- names(x)
          ans
      })

setMethod("length",
          signature(x = "indirectExternalMatrix"),
          function (x)
      {
          length(x@index)
      })

setReplaceMethod("length",
                 signature(x = "indirectExternalMatrix"),
                 function (x, value)
             {
                 stop("cannot modify length of indirect external matrix")
             })

setMethod("dim", "indirectExternalMatrix",
          function(x) dim(x@index))

setReplaceMethod("dim", signature(x="indirectExternalMatrix",
                                  value="NULL"),
                 function(x, value)
             {
                 new("indirectExternalVector",
                     actual=x@actual,
                     index=as(x@index, "vectorIndex"))
             })

setReplaceMethod("dim", signature(x="indirectExternalMatrix"),
                 function(x, value)
             {
                 dim(x@index) <- value
                 x
             })

setMethod("names",
          signature(x="indirectExternalMatrix"),
          function (x)
      {
          names(x@index)
      })

setReplaceMethod("names",
          signature(x="indirectExternalMatrix"),
          function(x, value)
      {
          names(x@index) <- value
          x
      })

setMethod("dimnames",
          signature(x="indirectExternalMatrix"),
          function (x)
      {
          dimnames(x@index)
      })

setReplaceMethod("dimnames",
                 signature(x="indirectExternalMatrix"),
                 function(x, value)
             {
                 dimnames(x@index) <- value
                 x
             })

setMethod("[",
          signature(x = "indirectExternalMatrix", i = "missing",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop)
      {
          if (nargs() < 3)
              new("indirectExternalVector",
                  actual=x@actual,
                  index=as(x@index, "vectorIndex"))
          else x
      })

setMethod("[",
          signature(x = "indirectExternalMatrix", i = "missing",
                    j = "missing", drop = "logical"),
          function (x, i, j, ..., drop)
      {
          if (nargs() < 4)
              new("indirectExternalVector",
                  actual=x@actual,
                  index=as(x@index, "vectorIndex"))
          else if (drop) {
              i <- x@index[,,drop=TRUE]
              if (is(i, 'matrixIndex')) {
                  x@index <- i
                  x
              } else {
                  new("indirectExternalVector",
                      actual=x@actual,
                      index=i)
              }
          } else x
      })

setMethod("[",
          signature(x = "indirectExternalMatrix",
                    i = "missing", drop = "missing"),
          function (x, i, j, ..., drop)
      {
          if (nargs() != 3)
              stop('incorrect number of subscripts')
          callGeneric(x, new("allIndex"), j, drop=FALSE)
      })

setMethod("[",
          signature(x = "indirectExternalMatrix",
                    i = "missing", drop = "logical"),
          function (x, i, j, ..., drop)
      {
          if (nargs() != 4)
              stop('incorrect number of subscripts')
          callGeneric(x, new("allIndex"), j, drop=drop)
      })

setMethod("[",
          signature(x = "indirectExternalMatrix",
                    j = "missing", drop = "missing"),
          function (x, i, j, ..., drop)
      {
          if (nargs() == 2)
              callGeneric(x, i, drop=FALSE)
          else if (nargs() == 3)
              callGeneric(x, i, new("allIndex"), drop=FALSE)
          else stop('incorrect number of subscripts')
      })

setMethod("[",
          signature(x = "indirectExternalMatrix",
                    j = "missing", drop = "logical"),
          function (x, i, j, ..., drop)
      {
          if (nargs() == 4)
              callGeneric(x, i, new("allIndex"), drop=drop)
          else if (nargs() == 3) {
              i <- x@index[i]
              if (length(i) == 0)
                  force(x@actual)[i]
              else if (anyNA(i))
                  stop('subscript out of bounds')
              else {
                  ans <- new("indirectExternalVector",
                             actual=x@actual,
                             index=i)
                  if (!is.null(names(x)))
                      names(ans) <- Subset(names(x), i)
                  ans
              }
          } else stop('incorrect number of subscripts')
      })

setMethod("[",
          signature(x = "indirectExternalMatrix",
                    drop = "missing"),
          function (x, i, j, ..., drop)
      {
          if (nargs() != 3)
              stop('incorrect number of subscripts')
          callGeneric(x, i, j, drop=FALSE)
      })

setMethod("[",
          signature(x = "indirectExternalMatrix"),
          function (x, i, j, ..., drop)
      {
          if (nargs() != 4)
              stop('incorrect number of subscripts')
          i <- x@index[i, j, drop = drop]
          if (length(i) == 0) {
              if (is(i, "matrixIndex"))
                  force(x@actual)[i[[1]], i[[2]], drop=drop]
              else force(x@actual)[i, drop=drop]
          } else if (anyNA(i))
              stop('subscript out of bounds')
          else if (is(i, 'matrixIndex')) {
              x@index <- i
              x
          } else {
              new("indirectExternalVector",
                  actual=x@actual,
                  index=i)
          }
      })

setReplaceMethod("[",
                 signature(x = "indirectExternalMatrix", i = "missing",
                           j = "missing"),
                 function (x, i, j, ..., value)
             {
                 if (nargs() > 4)
                     stop('incorrect number of subscripts')
                 if (nargs() == 4)
                     x@actual[x@index[[1]], x@index[[2]]] <- value
                 else x@actual[x@index] <- value
                 x
             })

setReplaceMethod("[",
                 signature(x = "indirectExternalMatrix"),
                 function (x, i, j, ..., value)
             {
                 nsubs <- nargs() - 2
                 if (nsubs == 0)
                     nsubs <- 1
                 if (nsubs == 1) {
                     i <- if (missing(i)) x@index else x@index[i]
                     if (anyNA(i))
                         stop('subscript out of bounds')
                     x@actual[i] <- value
                 } else if (nsubs == 2) {
                     i <- if (missing(i)) x@index[[1]] else x@index[[1]][i]
                     if (anyNA(i))
                         stop('subscript out of bounds')
                     j <- if (missing(j)) x@index[[2]] else x@index[[2]][j]
                     if (anyNA(j))
                         stop('subscript out of bounds')
                     x@actual[i, j] <- value
                 } else stop('incorrect number of subscripts')
                 x
             })

setAs("externalVector", "indirectExternalVector",
          function (from, to)
      {
          to <- new("indirectExternalVector", actual=from,
                    index = makeSlice(0, length(from), 1))
          if (!is.null(names(from)))
              names(to) <- names(from)
          to
      })

setAs("externalVector", "indirectExternalMatrix",
          function (from, to)
      {
          if (length(dim(from)) == 2) {
              to <- new("indirectExternalMatrix", actual=from,
                        index = new("matrixIndex",
                                    row=makeSlice(0, nrow(from), 1),
                                    col=makeSlice(0, ncol(from), 1),
                                    origdim=dim(from)))
              if (!is.null(dimnames(from)))
                  dimnames(to) <- dimnames(from)
              if (!is.null(names(from)))
                  names(to) <- names(from)
          } else {
              to <- new("indirectExternalVector", actual=from,
                        index = makeSlice(0, length(from), 1))
              if (!is.null(names(from)))
                  names(to) <- names(from)
          }
          to
      })

setAs( "indirectExternalVector", "vector",
      function(from, to) as.vector(from))

setAs( "indirectExternalVector", "matrix",
      function(from, to) as.matrix(from))

setAs( "indirectExternalMatrix", "vector",
      function(from, to) as.vector(from))

setAs( "indirectExternalMatrix", "matrix",
      function(from, to) as.matrix(from))

setAs("vector", "indirectExternalVector",
      function(from, to)
  {
      as(as(from, "externalVector"), "indirectExternalVector")
  })

setAs("vector", "indirectExternalMatrix",
      function(from, to)
  {
      as(externalMatrix(from), "indirectExternalMatrix")
  })

setAs("matrix", "indirectExternalMatrix",
      function(from, to)
  {
      as(as(from, "externalVector"), "indirectExternalMatrix")
  })

setMethod("Ops",
          signature(e2="indirectExternalVector"),
          function(e1, e2)
      {
          callGeneric(e1, force(e2))
      })

setMethod("Ops",
          signature(e1="indirectExternalVector"),
          function(e1, e2)
      {
          callGeneric(force(e1), e2)
      })

setMethod("Ops",
          signature(e1="indirectExternalVector",
                    e2="indirectExternalVector"),
          function(e1, e2)
      {
          callGeneric(force(e1), force(e2))
      })

setMethod("Complex",
          signature(z="indirectExternalVector"),
          function(z)
      {
          callGeneric(force(z))
      })

setMethod("Math",
          signature(x="indirectExternalVector"),
          function(x)
      {
          callGeneric(force(x))
      })

setMethod("Ops",
          signature(e2="indirectExternalMatrix"),
          function(e1, e2)
      {
          callGeneric(e1, force(e2))
      })

setMethod("Ops",
          signature(e1="indirectExternalMatrix"),
          function(e1, e2)
      {
          callGeneric(force(e1), e2)
      })

setMethod("Ops",
          signature(e1="indirectExternalMatrix",
                    e2="indirectExternalMatrix"),
          function(e1, e2)
      {
          callGeneric(force(e1), force(e2))
      })

setMethod("Ops",
          signature(e1="indirectExternalMatrix",
                    e2="indirectExternalVector"),
          function(e1, e2)
      {
          callGeneric(force(e1), force(e2))
      })

setMethod("Ops",
          signature(e1="indirectExternalVector",
                    e2="indirectExternalMatrix"),
          function(e1, e2)
      {
          callGeneric(force(e1), force(e2))
      })

setMethod("Complex",
          signature(z="indirectExternalMatrix"),
          function(z)
      {
          callGeneric(force(z))
      })

setMethod("Math",
          signature(x="indirectExternalMatrix"),
          function(x)
      {
          callGeneric(force(x))
      })

setMethod("log",
          signature(x="indirectExternalVector"),
          function(x, base = exp(1))
      {
          callGeneric(force(x), if(class(base) %in% c("indirectExternalVector", "indirectExternalMatrix")) force(base) else base)
      })

setMethod("log",
          signature(x="indirectExternalMatrix"),
          function(x, base = exp(1))
      {
          callGeneric(force(x), if(class(base) %in% c("indirectExternalVector", "indirectExternalMatrix")) force(base) else base)
      })

setMethod("is.finite",
          signature(x = "indirectExternalMatrix"),
          function (x) is.finite(force(x)))

setMethod("is.infinite",
          signature(x = "indirectExternalMatrix"),
          function (x) is.infinite(force(x)))

setMethod("is.na",
          signature(x = "indirectExternalMatrix"),
          function (x) is.na(force(x)))

setMethod("is.nan",
          signature(x = "indirectExternalMatrix"),
          function (x) is.nan(force(x)))

setMethod("anyNA",
          signature(x = "indirectExternalMatrix"),
          function (x) anyNA(x@index) || anyNA(force(x)))

setMethod("allNA",
          signature(x = "indirectExternalMatrix"),
          function (x) allNA(force(x)))

setMethod("is.finite",
          signature(x = "indirectExternalVector"),
          function (x) is.finite(force(x)))

setMethod("is.infinite",
          signature(x = "indirectExternalVector"),
          function (x) is.infinite(force(x)))

setMethod("is.na",
          signature(x = "indirectExternalVector"),
          function (x) is.na(force(x)))

setMethod("is.nan",
          signature(x = "indirectExternalVector"),
          function (x) is.nan(force(x)))

setMethod("anyNA",
          signature(x = "indirectExternalVector"),
          function (x) anyNA(x@index) || anyNA(force(x)))

setMethod("allNA",
          signature(x = "indirectExternalVector"),
          function (x) allNA(force(x)))
