externalVector <-
    function(type = "logical", length = 0, storageClass =
    "simpleStorage", indirect = FALSE)
{
    if (!is.character(type))
        stop('invalid argument type = ', type)
    type <-
        if (type %in% c("logical", "numeric", "integer", "complex",
                        "character"))
            vector(type, 1)
        else new(type)
    ans <- new("externalVectorWithStorage",
               storage = new(storageClass, type = type,
               length = as.integer(length)))
    if (indirect)
        as(ans, "indirectExternalVector")
    else ans
}

externalNumeric <-
    function(length = 0, storageClass = "simpleStorage", indirect = FALSE)
    externalVector(type = "numeric", length, storageClass, indirect)

externalInteger <-
    function(length = 0, storageClass = "simpleStorage", indirect = FALSE)
    externalVector(type = "integer", length, storageClass, indirect)

externalLogical <-
    function(length = 0, storageClass = "simpleStorage", indirect = FALSE)
    externalVector(type = "logical", length, storageClass, indirect)

externalComplex <-
    function(length = 0, storageClass = "simpleStorage", indirect = FALSE)
    externalVector(type = "complex", length, storageClass, indirect)

externalCharacter <-
    function(length = 0, storageClass = "simpleStorage", indirect = FALSE)
    externalVector(type = "character", length, storageClass, indirect)

externalMatrix <-
    function (data = NA, nrow = 1, ncol = 1, byrow = FALSE,
              dimnames = NULL,
              storageClass = (if (is(data, "externalVectorWithStorage"))
                              class(data@storage) else
                              "simpleStorage"),
              indirect = FALSE)
{
    if (!is(data, "externalVectorWithStorage"))
        data <- as.Rvector(data)
    ansTypeName <- class(defaultElement(data))
    if (missing(nrow)) 
        nrow <- ceiling(length(data)/ncol)
    else if (missing(ncol)) 
        ncol <- ceiling(length(data)/nrow)
    anslen <- nrow*ncol
    if (byrow && anslen > 1)
        stop("option `byrow = TRUE' not implemented")
    datlen <- length(data)
    if (length(data) < 0)
        stop("argument has length zero")
    ## The following is mostly copied from do_matrix in R/src/main/array.c
    if (datlen > 1 && anslen %% datlen != 0) {
	if (((datlen > nrow) && (datlen / nrow) * nrow != datlen) ||
	    ((datlen < nrow) && (nrow / datlen) * datlen != nrow))
	    warning("Replacement length not a multiple of the elements to replace in externalMatrix(...)")
	else if (((datlen > ncol) && (datlen / ncol) * ncol != datlen) ||
		 ((datlen < ncol) && (ncol / datlen) * datlen != ncol))
	    warning("Replacement length not a multiple of the elements to replace in matrix(...)")
    } else if ((datlen > 1) && (anslen == 0)){
        warning("Replacement length not a multiple of the elements to replace in matrix(...)")
    } else if (datlen == 0 && anslen > 0){
        stop("No data to replace in externalMatrix(...)")
    }
    ## now we create the matrix and copy the values to it
    ans <- externalVector(ansTypeName, anslen, storageClass,
                          indirect=indirect)
    ans[] <- data
    dim(ans) <- c(nrow, ncol)
    dimnames(ans) <- dimnames
    ans
}
