#include "storageMethods.h"

typedef enum {COLSUM=0, COLMEAN=1, ROWSUM=2, ROWMEAN=3} optype_t;

static SEXP
colRowSums1(SEXP x, storageMethods_t* methods, getExtMatrixElt_t getElt,
            int n, int p, Rboolean NaRm, optype_t OP)
{
    SEXP ans = R_NilValue;
    SEXP storage = GET_STORAGE_EXT_PTR(GET_SLOT(x, install("storage")));
    int cnt = 0, i, j;
    SEXPTYPE type = methods->mode;
    Rboolean keepNA;
    double sum = 0.0;
    primitiveElt_t elt;
    int warn = 0;

    keepNA = !NaRm;

    if (OP == 0 || OP == 1) { /* columns */
        cnt = n;
        PROTECT(ans = allocVector(REALSXP, p));
        for (j = 0; j < p; j++) {
            switch (type) {
            case REALSXP:
#ifdef IEEE_754
                if (keepNA)
                    for (sum = 0., i = 0; i < n; i++) {
                        getElt(storage, methods, i, j,
                               p, n, &elt, &warn);
                        sum += elt.numeric;
                    }
                else {
                    for (cnt = 0, sum = 0., i = 0; i < n; i++) {
                        getElt(storage, methods, i, j,
                               p, n, &elt, &warn);
                        if (!ISNAN(elt.numeric)) {cnt++; sum += elt.numeric;}
                    }
                }
#else
                for (cnt = 0, sum = 0., i = 0; i < n; i++, rx++) {
                    getElt(storage, methods, i, j,
                           p, n, &elt, &warn);
                    if (!ISNAN(elt.numeric)) {cnt++; sum += elt.numeric;}
                    else if (keepNA) {sum = NA_REAL; break;}
                }
#endif
                break;
            case INTSXP:
                for (cnt = 0, sum = 0., i = 0; i < n; i++) {
                    getElt(storage, methods, i, j,
                           p, n, &elt, &warn);
                    if (elt.integer != NA_INTEGER) {
                        cnt++;
                        sum += elt.integer;
                    } else if (keepNA) {sum = NA_REAL; break;}
                }
                break;
            case LGLSXP:
                for (cnt = 0, sum = 0., i = 0; i < n; i++) {
                    getElt(storage, methods, i, j,
                           p, n, &elt, &warn);
                    if (elt.logical != NA_LOGICAL) {
                        cnt++;
                        if (elt.logical)
                            sum++;
                    }
                    else if (keepNA) {sum = NA_REAL; break;}
                }
                break;
            }
            if (OP == 1) {
                if (!ISNAN(sum) && cnt > 0)
                    sum /= cnt;
            }
            REAL(ans)[j] = sum;
        }
    }

    if (OP == 2 || OP == 3) { /* rows */
        cnt = p;
        PROTECT(ans = allocVector(REALSXP, n));

#ifdef IEEE_754
        /* reverse summation order to improve cache hits */
        if (type == REALSXP) {
            double *rans = REAL(ans), *ra = rans, *cntvec = NULL, *c;
            if (!keepNA && OP == 3) cntvec = Calloc(n, double);
            for (ra = rans, i = 0; i < n; i++) *ra++ = 0.0;
            for (j = 0; j < p; j++) {
                ra = rans;
                if (keepNA)
                    for (i = 0; i < n; i++) {
                        getElt(storage, methods, i, j,
                               p, n, &elt, &warn);
                        *ra++ += elt.numeric;
                    }
                else
                    for (c = cntvec, i = 0; i < n; i++, ra++, c++) {
                        getElt(storage, methods, i, j,
                               p, n, &elt, &warn);
                        if (!ISNAN(elt.numeric)) {
                            *ra += elt.numeric;
                            if (OP == 3) (*c)++;
                        }
                    }
            }
            if (OP == 3) {
                if (keepNA)
                    for (ra = rans, i = 0; i < n; i++) 
                        *ra++ /= p;
                else {
                    for (ra = rans, c = cntvec, i = 0; i < n; i++, c++) 
                        if (!ISNAN(*ra) && *c > 0)
                            *ra++ /= *c;
                        else ra++;
                    Free(cntvec);
                }
            }
            UNPROTECT(1);
            return ans;
        }
#endif

        for (i = 0; i < n; i++) {
            switch (type) {
#ifndef IEEE_754
            case REALSXP:
                for (cnt = 0, sum = 0., j = 0; j < p; j++) {
                    getElt(storage, methods, i, j,
                           p, n, &elt, &warn);
                    if (!ISNAN(elt.numeric)) {cnt++; sum += elt.numeric;}
                    else if (keepNA) {sum = NA_REAL; break;}
                }
                break;
#endif
            case INTSXP:
                for (cnt = 0, sum = 0., j = 0; j < p; j++) {
                    getElt(storage, methods, i, j,
                           p, n, &elt, &warn);
                    if (elt.integer != NA_INTEGER) {cnt++; sum += elt.integer;}
                    else if (keepNA) {sum = NA_REAL; break;}
                }
                break;
            case LGLSXP:
                for (cnt = 0, sum = 0., j = 0; j < p; j++) {
                    getElt(storage, methods, i, j,
                           p, n, &elt, &warn);
                    if (elt.logical != NA_LOGICAL) {
                        cnt++;
                        if (elt.logical)
                            sum++;
                    } else if (keepNA) {sum = NA_REAL; break;}
                }
                break;
            }
            if (OP == 3) {
                if (!ISNAN(sum) && cnt > 0)
                    sum /= cnt;
            }
            REAL(ans)[i] = sum;
        }
    }

    UNPROTECT(1);
    return ans;
}

static SEXP
colRowSums(SEXP x, storageMethods_t* methods,
           int n, int p, Rboolean NaRm, optype_t OP)
{
    getExtMatrixElt_t getElt;
    SEXPTYPE type;

    if (n == NA_INTEGER || n <= 0)
        error("invalid value of n");
    if (p == NA_INTEGER || p <= 0)
        error("invalid value of p");
    if (NaRm == NA_LOGICAL)
        error("invalid value of na.rm");

    switch (type = methods->mode) {
    case LGLSXP:
    case INTSXP:
    case REALSXP:
        getElt = methods->getMatrixElt;
        break;
    case CPLXSXP:
        getElt = methods->imaginaryGetMatrixElt;
        break;
    default:
        error("`x' must be numeric");
        return R_NilValue; /* -Wall */
    }
    return colRowSums1(x, methods, getElt, n, p, NaRm, OP);
}

static SEXP
colSums(SEXP x, SEXP xmethods, SEXP n, SEXP p, SEXP NaRm)
{
    return colRowSums(x, (storageMethods_t*) R_ExternalPtrAddr(xmethods),
                      asInteger(n), asInteger(p), asLogical(NaRm), COLSUM);
}

static SEXP
colMeans(SEXP x, SEXP xmethods, SEXP n, SEXP p, SEXP NaRm)
{
    return colRowSums(x, (storageMethods_t*) R_ExternalPtrAddr(xmethods),
                      asInteger(n), asInteger(p), asLogical(NaRm), COLMEAN);
}

static SEXP
rowSums(SEXP x, SEXP xmethods, SEXP n, SEXP p, SEXP NaRm)
{
    return colRowSums(x, (storageMethods_t*) R_ExternalPtrAddr(xmethods),
                      asInteger(n), asInteger(p), asLogical(NaRm), ROWSUM);
}

static SEXP
rowMeans(SEXP x, SEXP xmethods, SEXP n, SEXP p, SEXP NaRm)
{
    return colRowSums(x, (storageMethods_t*) R_ExternalPtrAddr(xmethods),
                      asInteger(n), asInteger(p), asLogical(NaRm), ROWMEAN);
}


#include <R.h>
#include <R_ext/Rdynload.h>

static const R_CallMethodDef R_CallDef  [] = {
    CALL_DEF(colSums, 5),
    CALL_DEF(rowSums, 5),
    CALL_DEF(colMeans, 5),
    CALL_DEF(rowMeans, 5),
    {NULL, NULL, 0},
};

const R_CallMethodDef*
array_Call_methods(int* nptr)
{
    *nptr = LENGTH_OF_METHODS_ARRAY(R_CallDef);
    return &R_CallDef[0];
}
