# include "common.h"

#ifdef elt
#undef elt
#endif


SEXP gcExternalPtr_resizeInternal(SEXP ptr, int size, int copy);

typedef union {
    double numeric;
    int integer;
    Rboolean logical;
    Rcomplex cmplx;
    SEXP character; /* a CHARSXP */
    SEXP any;
} primitiveElt_t;

typedef struct storageMethods storageMethods_t;

SEXP getStorageExternalPtr(SEXP x);

#define GET_STORAGE_EXT_PTR(x) getStorageExternalPtr(x)
#define GET_STORAGE_EXT_PTR_UNCHECKED(x) GET_SLOT((SEXP) (x), install("ptr"))
#define SET_STORAGE_EXT_PTR(x, v) SET_SLOT((SEXP) (x), install("ptr"), (v))

typedef primitiveElt_t* (*getExtElt_t)(const SEXPREC* storage,
                                       const storageMethods_t* methods, int i,
                                       primitiveElt_t* val, int*);

typedef const primitiveElt_t* (*setExtElt_t)(SEXP storage,
                                             const storageMethods_t* methods,
                                             int i,
                                             const primitiveElt_t* val, int*);

typedef primitiveElt_t*
(*getExtMatrixElt_t)(const SEXPREC* storage,
                     const storageMethods_t*, int i, int j,
                     int m, int n,
                     primitiveElt_t* value, int*);

typedef const primitiveElt_t*
(*setExtMatrixElt_t)(SEXP storage,
                     const storageMethods_t* methods, int i, int j,
                     int m, int n,
                     const primitiveElt_t* value, int*);

/*
 *  Get the subset corresponding to poitive integers in i.
 *  Returns the subset as an R vector of the same mode as the vector
 *  stored in the external pointer and of same length as i.
 *  
 *  If any of the indices in i is greater than the length of the
 *  external vector, the corresponding element in the returned value
 *  should be NA of appropriate type.
 */
typedef SEXP (*extSubset_t)(const SEXPREC* storage, storageMethods_t* methods,
                            const SEXPREC* i, int nx, SEXP value, int*);

/*
 *  Assign values corresponding to poitive integers in i.
 *  Length of i and value are same. Mode of value and the stored
 *  vector are also same. Any NA_INTEGER in i should be ignored.
 *
 *  Returns value.
 */
typedef const SEXPREC* (*extSubassign_t)(SEXP storage,
                                         storageMethods_t* methods,
                                         const SEXPREC* i, int nx,
                                         const SEXPREC* value, int*);

typedef SEXP
(*extMatrixSubset_t)(const SEXPREC* storage, storageMethods_t* methods,
                     const SEXPREC* s1,
                     const SEXPREC* s2, int m, int n,
                     SEXP value, int*);
typedef const SEXPREC*
(*extMatrixSubassign_t)(SEXP storage, storageMethods_t* methods,
                        const SEXPREC* s1,
                        const SEXPREC* s2, int m, int n,
                        const SEXPREC* value, int*);

typedef primitiveElt_t*
(*getRElt_t)(const SEXPREC* x, int i, primitiveElt_t* val);
               
typedef const primitiveElt_t*
(*setRElt_t)(SEXP x, int i, const primitiveElt_t* val);
               
struct storageMethods {
    getRElt_t logicalGetRElt;
    getRElt_t integerGetRElt;
    getRElt_t numericGetRElt;
    getRElt_t complexGetRElt;
    getRElt_t characterGetRElt;
    getRElt_t listGetRElt;
    getRElt_t getRElt;
    setRElt_t logicalSetRElt;
    setRElt_t integerSetRElt;
    setRElt_t numericSetRElt;
    setRElt_t complexSetRElt;
    setRElt_t characterSetRElt;
    setRElt_t listSetRElt;
    setRElt_t setRElt;

    /* get the i-th element */
    getExtElt_t numericGetElt;
    getExtElt_t integerGetElt;
    getExtElt_t logicalGetElt;
    getExtElt_t complexGetElt;
    getExtElt_t imaginaryGetElt;
    getExtElt_t characterGetElt;
    getExtElt_t anyGetElt;
    getExtElt_t getExtElt;
    /* set the i-th element */
    setExtElt_t numericSetElt;
    setExtElt_t integerSetElt;
    setExtElt_t logicalSetElt;
    setExtElt_t complexSetElt;
    setExtElt_t imaginarySetElt;
    setExtElt_t characterSetElt;
    setExtElt_t anySetElt;
    setExtElt_t setExtElt;

    extSubset_t numericSubset;
    extSubset_t integerSubset;
    extSubset_t logicalSubset;
    extSubset_t complexSubset;
    extSubset_t characterSubset;
    extSubset_t anySubset;
    extSubset_t vectorSubset;

    extSubassign_t numericSubassign;
    extSubassign_t integerSubassign;
    extSubassign_t logicalSubassign;
    extSubassign_t complexSubassign;
    extSubassign_t characterSubassign;
    extSubassign_t anySubassign;

    getExtMatrixElt_t numericGetMatrixElt;
    getExtMatrixElt_t integerGetMatrixElt;
    getExtMatrixElt_t logicalGetMatrixElt;
    getExtMatrixElt_t complexGetMatrixElt;
    getExtMatrixElt_t imaginaryGetMatrixElt;
    getExtMatrixElt_t characterGetMatrixElt;
    getExtMatrixElt_t anyGetMatrixElt;
    getExtMatrixElt_t getMatrixElt;

    setExtMatrixElt_t numericSetMatrixElt;
    setExtMatrixElt_t integerSetMatrixElt;
    setExtMatrixElt_t logicalSetMatrixElt;
    setExtMatrixElt_t complexSetMatrixElt;
    setExtMatrixElt_t imaginarySetMatrixElt;
    setExtMatrixElt_t characterSetMatrixElt;
    setExtMatrixElt_t anySetMatrixElt;
    setExtMatrixElt_t setMatrixElt;

    extMatrixSubset_t numericMatrixSubset;
    extMatrixSubset_t integerMatrixSubset;
    extMatrixSubset_t logicalMatrixSubset;
    extMatrixSubset_t complexMatrixSubset;
    extMatrixSubset_t characterMatrixSubset;
    extMatrixSubset_t anyMatrixSubset;
    extMatrixSubset_t matrixSubset;

    extMatrixSubassign_t numericMatrixSubassign;
    extMatrixSubassign_t integerMatrixSubassign;
    extMatrixSubassign_t logicalMatrixSubassign;
    extMatrixSubassign_t complexMatrixSubassign;
    extMatrixSubassign_t characterMatrixSubassign;
    extMatrixSubassign_t anyMatrixSubassign;

    SEXP (*alloc)(const SEXPREC* x, SEXP type, int len);
    int (*size)(const SEXPREC* x);
    SEXP (*resize)(SEXP x, int len);
    /*
     *   Here mode is one of INTSXP, REALSXP, LGLSXP, CPLXSXP or
     *   STRSXP. Should use SET_EXT_RAWSTORE and SET_EXT_R_STORE to
     *   set any allocated storage in x.
     */
    SEXPTYPE mode;

    primitiveElt_t NA;
};

Rboolean isExternalVectorWithStorage(SEXP x);
storageMethods_t* getStorageMethods(SEXP x);
SEXP getArithmeticAnswerStorage(SEXP xstorage,
                                const storageMethods_t* xmethods,
                                int nx,
                                SEXP ystorage,
                                const storageMethods_t* ymethods,
                                int ny, SEXPTYPE mode);
